#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2021, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2021. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #

import numpy
from scipy.interpolate import interp2d
from scipy.optimize import curve_fit
from scipy import integrate

from oasys.widgets.abstract.benders.bender_data_to_plot import BenderDataToPlot

# -----------------------------------------------------------------

class DoubleRodBenderParameters:
    W1 = None
    L = None
    p = None
    q = None
    grazing_angle = None
    optimized_length = None

    E = None
    h = None
    r = None
    l = None

    R0 = None
    R0_max = False
    R0_min = None
    R0_fixed = None

    eta = None
    eta_max = False
    eta_min = None
    eta_fixed = None

    W2 = None
    W2_max = False
    W2_min = None
    W2_fixed = None

    n_fit_steps = None

    workspace_units_to_m = None
    workspace_units_to_mm = None

    x = None
    y = None
    figure_error = None

def calculate_bender_correction(bender_parameter : DoubleRodBenderParameters):
    workspace_units_to_m  = bender_parameter.workspace_units_to_m
    workspace_units_to_mm = bender_parameter.workspace_units_to_mm

    x             = bender_parameter.x
    y             = bender_parameter.y
    W1            = bender_parameter.W1
    L             = bender_parameter.L
    p             = bender_parameter.p
    q             = bender_parameter.q
    grazing_angle = bender_parameter.grazing_angle

    optimized_length = bender_parameter.optimized_length

    E = bender_parameter.E
    l = bender_parameter.l
    h = bender_parameter.h
    r = bender_parameter.r

    if optimized_length is None:
        y_fit = y
    else:
        cursor = numpy.where(numpy.logical_and(y >= -optimized_length / 2, y <= optimized_length / 2))
        y_fit = y[cursor]

    ideal_slope_profile_fit = ideal_slope_profile(y_fit, p, q, grazing_angle)

    epsilon_minus = 1 - 1e-8
    epsilon_plus  = 1 + 1e-8

    initial_guess    = [ bender_parameter.R0, bender_parameter.eta, bender_parameter.W2]
    constraints     =  [[bender_parameter.R0_min if bender_parameter.R0_fixed == False else (bender_parameter.R0 * epsilon_minus),
                         bender_parameter.eta_min if bender_parameter.eta_fixed == False else (bender_parameter.eta * epsilon_minus),
                         bender_parameter.W2_min if bender_parameter.W2_fixed == False else (bender_parameter.W2 * epsilon_minus)],
                        [bender_parameter.R0_max if bender_parameter.R0_fixed == False else (bender_parameter.R0 * epsilon_plus),
                         bender_parameter.eta_max if bender_parameter.eta_fixed == False else (bender_parameter.eta * epsilon_plus),
                         bender_parameter.W2_max if bender_parameter.W2_fixed == False else (bender_parameter.W2 * epsilon_plus)]
                       ]

    def bender_function(x, R0, eta, W2):
        return __bender_slope_profile(x, p, q, grazing_angle, W1, L, R0 / workspace_units_to_m, eta, W2 / workspace_units_to_mm)

    for i in range(bender_parameter.n_fit_steps):
        parameters, _ = curve_fit(f=bender_function,
                                  xdata=y_fit,
                                  ydata=ideal_slope_profile_fit,
                                  p0=initial_guess,
                                  bounds=constraints,
                                  method='trf')
        initial_guess = parameters

    R0  = parameters[0] / workspace_units_to_m # here in workspace units
    eta = parameters[1]
    W2  = parameters[2] / workspace_units_to_mm

    alpha = calculate_taper_factor(W1, W2, L, p, q, grazing_angle)
    W0    = calculate_W0(W1, alpha, L, p, q, grazing_angle) # W at the center

    bender_profile = __bender_height_profile(y, p, q, grazing_angle, R0, eta, alpha)

    F_upstream, F_downstream = calculate_bender_forces(q, R0, eta, E, W0, l, h, r)

    parameters = numpy.append(parameters, round(alpha, 3))
    parameters = numpy.append(parameters, round(W0 * workspace_units_to_mm, 4))
    parameters = numpy.append(parameters, round(F_upstream, 6))
    parameters = numpy.append(parameters, round(F_downstream, 6))

    ideal_profile  = ideal_height_profile(y, p, q, grazing_angle)

    # back to Shadow system
    bender_profile -= numpy.min(bender_profile)
    ideal_profile  -= numpy.min(ideal_profile)

    # from here it's Shadow Axis system
    correction_profile = ideal_profile - bender_profile
    if not optimized_length is None: correction_profile_fit = correction_profile[cursor]

    # r-squared = 1 - residual sum of squares / total sum of squares
    r_squared = 1 - (numpy.sum(correction_profile ** 2) / numpy.sum((ideal_profile - numpy.mean(ideal_profile)) ** 2))
    rms       = round(correction_profile.std() * 1e9 * workspace_units_to_m, 6)
    if not bender_parameter.optimized_length is None: rms_opt = round(correction_profile_fit.std() * 1e9 * workspace_units_to_m, 6)

    z_bender_correction = numpy.zeros((len(x), len(y)))

    for i in range(z_bender_correction.shape[0]): z_bender_correction[i, :] = numpy.copy(correction_profile)

    bender_data_to_plot = BenderDataToPlot(x=x,
                                           y=y,
                                           ideal_profile=ideal_profile,
                                           bender_profile=bender_profile,
                                           correction_profile=correction_profile,
                                           titles=["Bender vs. Ideal Profiles" + "\n" + r'$R^2$ = ' + str(r_squared),
                                                   "Correction Profile 1D, r.m.s. = " + str(rms) + " nm" + ("" if optimized_length is None else (", " + str(rms_opt) + " nm (optimized)"))],
                                           z_bender_correction_no_figure_error=z_bender_correction)

    if not bender_parameter.figure_error is None:
        x_e, y_e, z_e = bender_parameter.figure_error

        if len(x) == len(x_e) and len(y) == len(y_e) and \
                x[0] == x_e[0] and x[-1] == x_e[-1] and \
                y[0] == y_e[0] and y[-1] == y_e[-1]:
            z_figure_error = z_e
        else:
            z_figure_error = interp2d(y_e, x_e, z_e, kind='cubic')(y, x)

        bender_data_to_plot.z_figure_error      = z_figure_error
        bender_data_to_plot.z_bender_correction = bender_data_to_plot.z_bender_correction_no_figure_error + z_figure_error
    else:
        bender_data_to_plot.z_bender_correction = bender_data_to_plot.z_bender_correction_no_figure_error

    return parameters, bender_data_to_plot

def __focal_distance(p, q):
    return p*q/(p+q)

def __demagnification_factor(p, q):
    return p/q

def __mu_nu(m):
    return (m - 1) / (m + 1), m/(m+1)**2

def __calculate_ideal_slope_variation(y, fprime, K0id, mu, nu):
    return 2*fprime*K0id*((2 * nu * (y / fprime) + mu) / numpy.sqrt(1 - mu * (y / fprime) - nu * (y / fprime) ** 2) - mu)

def __fprime(p, q, grazing_angle):
    return __focal_distance(p, q) / numpy.cos(grazing_angle)


def __calculate_bender_slope_variation(y, fprime, K0, eta, alpha):
    return -(K0*fprime/alpha**2)*(eta * alpha * (y / fprime) + (eta + alpha) * numpy.log(1 - (alpha * y / fprime)))

def calculate_taper_factor(W1, W2, L, p, q, grazing_angle):
    # W2 = W1(1 - alpha L/f')
    # W2/W1 - 1 = - alpha  L/f'
    # f'/L ( 1 - W2/W1) = alpha
    return (1 - W2/W1) * (__fprime(p, q, grazing_angle) / L)

def calculate_W0(W1, alpha, L, p, q, grazing_angle):
    return W1*(1 - alpha * L / (2 * __fprime(p, q, grazing_angle)))

def ideal_slope_profile(y, p, q, grazing_angle):
    mu, nu = __mu_nu(__demagnification_factor(p, q))
    fprime = __focal_distance(p, q)/numpy.cos(grazing_angle)
    K0id   = numpy.tan(grazing_angle)/(2*fprime)

    return __calculate_ideal_slope_variation(y, fprime, K0id, mu, nu)

def ideal_height_profile(y, p, q, grazing_angle):
    mu, nu = __mu_nu(__demagnification_factor(p, q))
    fprime = __focal_distance(p, q)/numpy.cos(grazing_angle)
    K0id   = numpy.tan(grazing_angle)/(2*fprime)

    profile = numpy.zeros(len(y))
    for i in range(len(y)):
        profile[i] = integrate.quad(func=(lambda x: __calculate_ideal_slope_variation(x, fprime, K0id, mu, nu)), a=y[0], b=y[i])[0]

    return profile

def __bender_slope_profile(y, p, q, grazing_angle, W1, L, R0, eta, W2):
    return __calculate_bender_slope_variation(y, __fprime(p, q, grazing_angle), 1 / R0, eta, alpha=calculate_taper_factor(W1, W2, L, p, q, grazing_angle))

def __bender_height_profile(y, p, q, grazing_angle, R0, eta, alpha):
    fprime = __fprime(p, q, grazing_angle)

    profile = numpy.zeros(len(y))
    for i in range(len(y)):
        profile[i] = integrate.quad(func=(lambda x: __calculate_bender_slope_variation(x, fprime, 1/R0, eta, alpha)), a=y[0], b=y[i])[0]

    return profile

# -----------------------------------------------
# q = focus distance (from mirror center) (1/p + 1/q = 1/f lenses equation)
# eta = bender asymmetry factor (from slope minimization)
# K0 = 1/R (radius of curvature at the center)
# E0 = Young's modulus
# L = length of the mirror
# r = distance between inner ond outer rods
# ----------------------------------------------------------------
def calculate_bender_forces(q, R0, eta, E, W0, L, h, r):
    I0 = (W0*h**3)/12
    M0 = E*I0/R0
    F_upstream = (M0/r) * (1 - (eta*(L + 2*r)/(2*q)))
    F_downstream = (M0/r) * (1 + (eta*(L + 2*r)/(2*q)))

    return F_upstream, F_downstream





# THIS FILE IS GENERATED FROM OASYS SETUP.PY
short_version = '1.2.127'
version = '1.2.127'
full_version = '1.2.127'
git_revision = '66ac370081054401fdfc9b51989c20e3900a03e1'
release = True

if not release:
    version = full_version
    short_version += ".dev"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2021, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2021. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
import numpy

from oasys.widgets import widget

from matplotlib.pyplot import rcParams
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

from matplotlib.figure import Figure
from oasys.widgets.gui import FigureCanvas3D
from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence

from orangewidget import gui
from orangewidget.settings import Setting

from PyQt5.QtCore import QRect
from PyQt5.Qt import QApplication, Qt
from PyQt5.QtWidgets import QMessageBox, QSlider

pixels_to_inches = 1/rcParams['figure.dpi']

class OpticalElementsColors:
    MIRROR = (0, 0, 1, 0.1)
    CRYSTAL = (0, 1, 0, 0.1)
    GRATING = (1, 0, 0, 0.1)
    SLITS = (165/255, 42/255, 42/255, 0.1)
    LENS = (128/255, 0/255, 128/255, 0.1)

class Orientations:
    UP = 0
    DOWN = 1
    LEFT = 2
    RIGHT = 3

class AspectRatioModifier:
    def __init__(self,
                 element_expansion_factor  = [1.0, 1.0, 1.0], # length,   width, thickness
                 layout_reduction_factor   = [1.0, 1.0, 1.0]  # distance, shift, height
                 ):
        self.element_expansion_factor  = element_expansion_factor
        self.layout_reduction_factor = layout_reduction_factor

DEFAULT_AZIM = -23.1
DEFAULT_ELEV = 27.9

class AbstractBeamlineRenderer(widget.OWWidget):
    want_main_area = 1

    WIDGET_WIDTH = 1900
    WIDGET_HEIGHT = 1000

    initial_height = Setting(0.0)
    use_axis = Setting(1)
    use_labels = Setting(1)
    draw_source = Setting(1)
    draw_optical_axis = Setting(1)

    element_expansion_factor =  Setting(1.0)
    distance_compression_factor = Setting(1.0)
    sagittal_expansion_factor = Setting(10.0)

    use_range = Setting(0)
    range_min = Setting(0.0)
    range_max = Setting(1.0)

    def __init__(self, is_using_workspace_units=False):
        super().__init__()

        self.is_using_workspace_units = is_using_workspace_units
        self.units_to_mm              = 1000

        self.build_layout()

    def build_layout(self):
        #self.warning(text="Widget under development: please do not report anomalies")

        geom = QApplication.desktop().availableGeometry()

        window_width = round(min(geom.width() * 0.98, self.WIDGET_WIDTH))
        window_height = round(min(geom.height() * 0.95, self.WIDGET_HEIGHT))
        control_area_width = 350

        self.setGeometry(QRect(round(geom.width() * 0.05),
                               round(geom.height() * 0.05),
                               window_width,
                               window_height))

        self.setMaximumHeight(self.geometry().height())
        self.setMaximumWidth(self.geometry().width())

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal", width=control_area_width)

        gui.button(button_box, self, "Restore Default View and Values", callback=self.reset_default, height=45)

        gen_box = oasysgui.widgetBox(self.controlArea, "Options", addSpace=False, orientation="vertical", width=control_area_width)

        beamline_box = oasysgui.widgetBox(gen_box, "Beamline", addSpace=False, orientation="vertical", height=170)

        gui.checkBox(beamline_box, self, "use_axis", "Show axis", labelWidth=200, callback=self.refresh)
        gui.checkBox(beamline_box, self, "use_labels", "Show labels", labelWidth=200, callback=self.refresh)
        gui.checkBox(beamline_box, self, "draw_source", "Draw source", labelWidth=200, callback=self.refresh)
        gui.checkBox(beamline_box, self, "draw_optical_axis", "Draw optical axis", labelWidth=200, callback=self.refresh)
        self.le_initial_height = oasysgui.lineEdit(beamline_box, self, "initial_height", "Beam vertical baseline ", labelWidth=230, valueType=float, orientation="horizontal", callback=self.refresh)

        oe_exp_box = oasysgui.widgetBox(gen_box, "Expansion/Compression (aesthetic)", addSpace=False, orientation="vertical", height=130)

        oasysgui.lineEdit(oe_exp_box, self, "element_expansion_factor", "O.E. expansion factor (\u22651)", labelWidth=230, valueType=float, orientation="horizontal", callback=self.refresh)
        oasysgui.lineEdit(oe_exp_box, self, "distance_compression_factor", "Length compression factor (\u22651)", labelWidth=230, valueType=float, orientation="horizontal", callback=self.refresh)
        oasysgui.lineEdit(oe_exp_box, self, "sagittal_expansion_factor", "Sagittal expansion factor (\u22651)", labelWidth=230, valueType=float, orientation="horizontal", callback=self.refresh)

        range_box = oasysgui.widgetBox(gen_box, "Range", addSpace=False, orientation="vertical", height=200)

        self.cb_use_range = gui.comboBox(range_box, self, "use_range", label="Use Layout Longitudinal Range", labelWidth=350,
                                         items=["No", "Yes"],
                                         callback=self.set_range, sendSelectedValue=False, orientation="horizontal")

        gui.separator(range_box, height=5)

        # - range min -----------------

        self.range_box_1 = oasysgui.widgetBox(range_box, "", addSpace=False, orientation="vertical")

        def range_min_value_change():
            self.slider_min.setValue(self.range_min*self.units_to_mm)
            self.refresh()

        self.le_range_min = oasysgui.lineEdit(self.range_box_1, self, "range_min", "Range Min ", labelWidth=220, valueType=float,
                                              orientation="horizontal", callback=range_min_value_change)

        slider_min_box = oasysgui.widgetBox(self.range_box_1, "", addSpace=False, orientation="horizontal")

        self.slider_min = QSlider(Qt.Horizontal)
        self.slider_min.setMinimum(0.0)
        self.slider_min.setMaximum(10.0)
        self.slider_min.setValue(0.0)
        self.slider_min.setTickPosition(QSlider.TicksBelow)
        self.slider_min.setTickInterval(1)
        self.slider_min.setFixedWidth(300)

        def slider_min_value_change():
            self.range_min = self.slider_min.value()/self.units_to_mm
            self.refresh()

        self.slider_min.valueChanged.connect(slider_min_value_change)

        slider_min_box.layout().addWidget(self.slider_min)

        # - range max -----------------

        def range_max_value_change():
            self.slider_max.setValue(self.range_max*self.units_to_mm)
            self.refresh()

        self.le_range_max = oasysgui.lineEdit(self.range_box_1, self, "range_max", "Range Max ", labelWidth=220, valueType=float,
                                              orientation="horizontal", callback=range_max_value_change)

        slider_max_box = oasysgui.widgetBox(self.range_box_1, "", addSpace=False, orientation="horizontal")

        self.slider_max = QSlider(Qt.Horizontal)
        self.slider_max.setMinimum(0.0)
        self.slider_max.setMaximum(10.0)
        self.slider_max.setValue(10.0)
        self.slider_max.setTickPosition(QSlider.TicksBelow)
        self.slider_max.setFixedWidth(300)
        self.slider_max.setTickInterval(1)

        def slider_max_value_change():
            self.range_max = self.slider_max.value()/self.units_to_mm
            self.refresh()

        self.slider_max.valueChanged.connect(slider_max_value_change)

        slider_max_box.layout().addWidget(self.slider_max)

        gui.rubber(self.controlArea)

        figure_width = window_width - control_area_width - 30
        figure_height = window_height - 20

        self.figure, self.axis = create_plot(figsize=(figure_width * pixels_to_inches, figure_height * pixels_to_inches))

        self.figure_canvas = FigureCanvas3D(fig=self.figure, ax=self.axis)
        self.figure_canvas.setFixedWidth(figure_width)
        self.figure_canvas.setFixedHeight(figure_height)

        self.mainArea.layout().addWidget(self.figure_canvas)

        gui.rubber(self.mainArea)

    def set_units_attributes(self, units_label, units_to_mm):
        self.units_label = units_label
        self.units_to_mm = units_to_mm
        self.tick_interval = 1

    def get_units_attributes(self):
        return "m", 1000

    def after_change_workspace_units(self):
        units_label, units_to_mm = self.get_units_attributes()

        label = self.le_initial_height.parent().layout().itemAt(0).widget()
        label.setText(label.text() + " [" + units_label + "]")
        label = self.le_range_min.parent().layout().itemAt(0).widget()
        label.setText(label.text() + " [" + units_label + "]")
        label = self.le_range_max.parent().layout().itemAt(0).widget()
        label.setText(label.text() + " [" + units_label + "]")

        self.set_units_attributes(units_label, units_to_mm)
        self.set_range(do_refresh=False)

    def refresh(self):
        self.render(reset_rotation=False)
        self.figure_canvas.draw()

    def set_range(self, do_refresh=True):
        if self.use_range == 1:
            self.range_box_1.setVisible(True)
            self.distance_compression_factor = 1.0
        else:
            self.range_box_1.setVisible(False)

        if do_refresh: self.refresh()

    def reset_default(self):
        self.element_expansion_factor = 1.0
        self.distance_compression_factor = 1.0
        self.use_range = 0
        self.set_range(do_refresh=False)
        self.reset_all()

    def reset_all(self):
        self.render(reset_rotation=True)
        self.figure_canvas.draw()

    def reset_rotation(self):
        self.axis.view_init(azim=DEFAULT_AZIM, elev=DEFAULT_ELEV)
        self.figure_canvas.draw()
        self.figure_canvas.mark_default_view()

    def check_fields(self, on_receiving_input=False):
        congruence.checkNumber(self.initial_height, "Beam vertical baseline")
        congruence.checkStrictlyPositiveNumber(self.element_expansion_factor, "O.E. expansion factor")
        congruence.checkStrictlyPositiveNumber(self.distance_compression_factor, "Layout compression factor")
        congruence.checkStrictlyPositiveNumber(self.sagittal_expansion_factor, "Sagittal expansion factor")
        if not on_receiving_input and self.use_range:
            if self.range_max <= self.range_min:
                self.range_max = self.range_min + self.tick_interval
                self.slider_max.setValue(self.range_max*self.units_to_mm)

    def render(self, reset_rotation=True, on_receiving_input=False):
        try:
            self.check_fields(on_receiving_input)

            render_result = self.render_beamline()

            if not render_result is None:
                number_of_elements, centers, limits = render_result

                limits[:, 0, :] *= self.sagittal_expansion_factor/self.element_expansion_factor # aestetic

                # sliders are integers, so they must be always in mm

                if on_receiving_input:
                    self.slider_min.setMinimum(numpy.min(limits[:, 1, :])*self.units_to_mm - self.tick_interval) # for visibility
                    self.slider_min.setMaximum(numpy.max(limits[:, 1, :])*self.units_to_mm - self.tick_interval) # for consistency
                    if self.range_min*self.units_to_mm == self.slider_min.minimum() + self.tick_interval:
                        self.range_min = self.slider_min.minimum()/self.units_to_mm
                        self.slider_min.setValue(self.slider_min.minimum())
                    else:
                        self.slider_min.setValue(max(self.range_min*self.units_to_mm, numpy.min(limits[:, 1, :])*self.units_to_mm))
                    self.slider_min.setTickInterval(self.tick_interval)

                    self.slider_max.setMinimum(numpy.min(limits[:, 1, :])*self.units_to_mm + self.tick_interval) # for consistency
                    self.slider_max.setMaximum(numpy.max(limits[:, 1, :])*self.units_to_mm + self.tick_interval) # for visibility
                    if self.range_max*self.units_to_mm == self.slider_max.maximum() - self.tick_interval:
                        self.range_max = self.slider_max.maximum()/self.units_to_mm
                        self.slider_max.setValue(self.slider_max.maximum())
                    else:
                        self.slider_max.setValue(min(self.range_max*self.units_to_mm, numpy.max(limits[:, 1, :])*self.units_to_mm))
                    self.slider_max.setTickInterval(self.tick_interval)

                if self.use_range == 1:
                    for i in range(number_of_elements): limits[i, 1, :] = numpy.array([self.range_min, self.range_max])

                if self.draw_optical_axis:
                    if self.use_range == 1: self.draw_central_radiation_line(centers=centers, rng=numpy.array([self.range_min, self.range_max]))
                    else:                   self.draw_central_radiation_line(centers=centers)

                self.format_axis(limits)

                if reset_rotation: self.reset_rotation()
                if on_receiving_input: self.figure_canvas.mark_default_view()

        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

    def render_beamline(self):
        raise NotImplementedError()

    def format_axis(self, limits):
        if self.use_axis:
            self.axis.set_axis_on()
            self.axis.set_ylabel("\n\n\n\n\nDistance along beam direction [" + self.units_label + "]")
            self.axis.axes.xaxis.set_ticklabels([])
            self.axis.axes.zaxis.set_ticklabels([])
            for line in self.axis.axes.xaxis.get_ticklines(): line.set_color(color='tab:gray')
            for line in self.axis.axes.zaxis.get_ticklines(): line.set_color(color='tab:gray')
            pane_color = (244/256,240/256,236/256, 1.0)
            self.axis.axes.xaxis.set_pane_color(color=pane_color)
            self.axis.axes.yaxis.set_pane_color(color=pane_color)
            self.axis.axes.zaxis.set_pane_color(color=pane_color)

            self.axis.set_xlim([numpy.min(limits[:, 0, :]), numpy.max(limits[:, 0, :])])
            self.axis.set_ylim([numpy.min(limits[:, 1, :]), numpy.max(limits[:, 1, :])])
            self.axis.set_zlim([numpy.min([0.0, numpy.min(limits[:, 2, :])]), numpy.max(limits[:, 2, :])])

            length_x = numpy.max(limits[:, 0, :]) - numpy.min(limits[:, 0, :])
            length_y = numpy.max(limits[:, 1, :]) - numpy.min(limits[:, 1, :])
            length_z = numpy.max(limits[:, 2, :])

            factor = numpy.max([length_x, length_y, length_z])

            self.axis.set_box_aspect(((length_x/factor), (length_y/factor), (length_z/factor)))

            self.axis.grid(True, color='tab:gray')
        else:
            self.axis.set_axis_off()

    def add_source(self, centers, limits, length=2.3, height=0.0, canting=0.0, aspect_ration_modifier=AspectRatioModifier(), source_name=None):
        length *= aspect_ration_modifier.element_expansion_factor[0]
        height *= aspect_ration_modifier.layout_reduction_factor[2]
        canting *= aspect_ration_modifier.layout_reduction_factor[0]
    
        center = numpy.array([0, canting, height])
    
        if canting != 0.0:
            self.axis.scatter(0, 0, height, s=30, c='c', marker='x')
            if self.use_labels: self.axis.text(0, 0, height, "Section Center")
    
        self.axis.scatter(center[0], center[1], center[2], s=30, c='r')
        if self.use_labels: self.axis.text(center[0], center[1], center[2],
                                           ("" if source_name is None else (source_name+" ")) + "Source Center")
    
        centers[0, :]   = center
        limits[0, 0, :] = numpy.array([0, 0])
        limits[0, 1, :] = numpy.array([canting - (length / 2), canting + (length / 2)])
        limits[0, 2, :] = numpy.array([0, height])

    def add_non_optical_element(self, centers, limits, oe_index, length,
                                distance, height, shift,
                                color=OpticalElementsColors.LENS, aspect_ration_modifier=AspectRatioModifier(), label=None):
        length *= aspect_ration_modifier.element_expansion_factor[0]
        distance *= aspect_ration_modifier.layout_reduction_factor[0]
        shift *= aspect_ration_modifier.layout_reduction_factor[1]
        height *= aspect_ration_modifier.layout_reduction_factor[2]

        half_width = 50 / self.units_to_mm
        half_length = length / 2
        half_thickness = half_width

        vertexes = [[-half_width, -half_length, -half_thickness],
                    [half_width, -half_length, -half_thickness],
                    [-half_width, -half_length, half_thickness],
                    [half_width, -half_length, half_thickness],
                    [-half_width, half_length, -half_thickness],
                    [half_width, half_length, -half_thickness],
                    [-half_width, half_length, half_thickness],
                    [half_width, half_length, half_thickness]
                    ]

        vertexes = numpy.array(vertexes)
        vertexes[:, 0] += shift
        vertexes[:, 1] += distance
        vertexes[:, 2] += height

        center = numpy.array([shift, distance, height])

        edges = [[vertexes[0], vertexes[1], vertexes[3], vertexes[2]],
                 [vertexes[0], vertexes[1], vertexes[5], vertexes[4]],
                 [vertexes[2], vertexes[3], vertexes[7], vertexes[6]],
                 [vertexes[0], vertexes[2], vertexes[6], vertexes[4]],
                 [vertexes[1], vertexes[3], vertexes[7], vertexes[5]],
                 [vertexes[4], vertexes[5], vertexes[7], vertexes[6]]
                 ]

        faces = Poly3DCollection(edges, linewidths=1, edgecolors='k')
        faces.set_facecolor(color)

        self.axis.add_collection3d(faces)

        self.axis.scatter(vertexes[:, 0], vertexes[:, 1], vertexes[:, 2], s=0)
        if label and self.use_labels: self.axis.text(center[0], center[1], center[2], label)

        centers[oe_index, :]   = center
        limits[oe_index, 0, :] = numpy.array([numpy.min(vertexes[:, 0]), numpy.max(vertexes[:, 0])])
        limits[oe_index, 1, :] = numpy.array([numpy.min(vertexes[:, 1]), numpy.max(vertexes[:, 1])])
        limits[oe_index, 2, :] = numpy.array([numpy.min(vertexes[:, 2]), numpy.max(vertexes[:, 2])])

    def add_optical_element(self, centers, limits, oe_index,
                            width, length, thickness, inclination,
                            distance, height, shift,
                            orientation=Orientations.UP, color=OpticalElementsColors.MIRROR,
                            aspect_ration_modifier=AspectRatioModifier(), label=None):
        length *= aspect_ration_modifier.element_expansion_factor[0]
        width *= aspect_ration_modifier.element_expansion_factor[1]
        thickness *= aspect_ration_modifier.element_expansion_factor[2]
        distance *= aspect_ration_modifier.layout_reduction_factor[0]
        shift *= aspect_ration_modifier.layout_reduction_factor[1]
        height *= aspect_ration_modifier.layout_reduction_factor[2]

        half_width = width / 2
        half_length = length / 2

        if orientation == Orientations.UP:
            vertexes = [[-half_width, -half_length * numpy.cos(inclination), -half_length * numpy.sin(inclination)],  # surface
                        [half_width, -half_length * numpy.cos(inclination), -half_length * numpy.sin(inclination)],  # surface
                        [-half_width, -(half_length * numpy.cos(inclination) - thickness * numpy.sin(inclination)), -(half_length * numpy.sin(inclination) + thickness * numpy.cos(inclination))],
                        [half_width, -(half_length * numpy.cos(inclination) - thickness * numpy.sin(inclination)), -(half_length * numpy.sin(inclination) + thickness * numpy.cos(inclination))],
                        [-half_width, half_length * numpy.cos(inclination), half_length * numpy.sin(inclination)],  # surface
                        [half_width, half_length * numpy.cos(inclination), half_length * numpy.sin(inclination)],  # surface
                        [-half_width, half_length * numpy.cos(inclination) + thickness * numpy.sin(inclination), half_length * numpy.sin(inclination) - thickness * numpy.cos(inclination)],
                        [half_width, half_length * numpy.cos(inclination) + thickness * numpy.sin(inclination), half_length * numpy.sin(inclination) - thickness * numpy.cos(inclination)]
                        ]
        elif orientation == Orientations.DOWN:
            vertexes = [[-half_width, -half_length * numpy.cos(inclination), -half_length * numpy.sin(inclination)],  # surface
                        [half_width, -half_length * numpy.cos(inclination), -half_length * numpy.sin(inclination)],  # surface
                        [-half_width, -(half_length * numpy.cos(inclination) + thickness * numpy.sin(inclination)), -half_length * numpy.sin(inclination) + thickness * numpy.cos(inclination)],
                        [half_width, -(half_length * numpy.cos(inclination) + thickness * numpy.sin(inclination)), -half_length * numpy.sin(inclination) + thickness * numpy.cos(inclination)],
                        [-half_width, half_length * numpy.cos(inclination), half_length * numpy.sin(inclination)],  # surface
                        [half_width, half_length * numpy.cos(inclination), half_length * numpy.sin(inclination)],  # surface
                        [-half_width, half_length * numpy.cos(inclination) - thickness * numpy.sin(inclination), half_length * numpy.sin(inclination) + thickness * numpy.cos(inclination)],
                        [half_width, half_length * numpy.cos(inclination) - thickness * numpy.sin(inclination), half_length * numpy.sin(inclination) + thickness * numpy.cos(inclination)]
                        ]
        elif orientation == Orientations.LEFT:
            vertexes = [[half_length * numpy.sin(inclination), -half_length * numpy.cos(inclination), -half_width],  # surface
                        [half_length * numpy.sin(inclination), -half_length * numpy.cos(inclination), half_width],  # surface
                        [half_length * numpy.sin(inclination) + thickness * numpy.cos(inclination), -(half_length * numpy.cos(inclination) - thickness * numpy.sin(inclination)), -half_width],
                        [half_length * numpy.sin(inclination) + thickness * numpy.cos(inclination), -(half_length * numpy.cos(inclination) - thickness * numpy.sin(inclination)), half_width],
                        [-half_length * numpy.sin(inclination), half_length * numpy.cos(inclination), -half_width],  # surface
                        [-half_length * numpy.sin(inclination), half_length * numpy.cos(inclination), half_width],  # surface
                        [-half_length * numpy.sin(inclination) + thickness * numpy.cos(inclination), half_length * numpy.cos(inclination) + thickness * numpy.sin(inclination), -half_width],
                        [-half_length * numpy.sin(inclination) + thickness * numpy.cos(inclination), half_length * numpy.cos(inclination) + thickness * numpy.sin(inclination), half_width]
                        ]
        elif orientation == Orientations.RIGHT:
            vertexes = [[half_length * numpy.sin(inclination), -half_length * numpy.cos(inclination), -half_width],  # surface
                        [half_length * numpy.sin(inclination), -half_length * numpy.cos(inclination), half_width],  # surface
                        [half_length * numpy.sin(inclination) - thickness * numpy.cos(inclination), -half_length * numpy.cos(inclination) - thickness * numpy.sin(inclination), -half_width],
                        [half_length * numpy.sin(inclination) - thickness * numpy.cos(inclination), -half_length * numpy.cos(inclination) - thickness * numpy.sin(inclination), half_width],
                        [-half_length * numpy.sin(inclination), half_length * numpy.cos(inclination), -half_width],  # surface
                        [-half_length * numpy.sin(inclination), half_length * numpy.cos(inclination), half_width],  # surface
                        [-half_length * numpy.sin(inclination) - thickness * numpy.cos(inclination), half_length * numpy.cos(inclination) - thickness * numpy.sin(inclination), -half_width],
                        [-half_length * numpy.sin(inclination) - thickness * numpy.cos(inclination), half_length * numpy.cos(inclination) - thickness * numpy.sin(inclination), half_width]
                        ]
    
        vertexes = numpy.array(vertexes)
        vertexes[:, 0] += shift
        vertexes[:, 1] += distance
        vertexes[:, 2] += height

        if orientation in [Orientations.UP, Orientations.DOWN]:
            center = numpy.array([(vertexes[1, 0] + vertexes[0, 0]) / 2, (vertexes[4, 1] + vertexes[0, 1]) / 2, (vertexes[4, 2] + vertexes[0, 2]) / 2])
        else:
            center = numpy.array([(vertexes[4, 0] + vertexes[0, 0]) / 2, (vertexes[4, 1] + vertexes[0, 1]) / 2, (vertexes[1, 2] + vertexes[0, 2]) / 2])

        edges = [[vertexes[0], vertexes[1], vertexes[3], vertexes[2]],
                 [vertexes[0], vertexes[1], vertexes[5], vertexes[4]],
                 [vertexes[2], vertexes[3], vertexes[7], vertexes[6]],
                 [vertexes[0], vertexes[2], vertexes[6], vertexes[4]],
                 [vertexes[1], vertexes[3], vertexes[7], vertexes[5]],
                 [vertexes[4], vertexes[5], vertexes[7], vertexes[6]]
                 ]
    
        faces = Poly3DCollection(edges, linewidths=1, edgecolors='k')
        faces.set_facecolor(color)
    
        self.axis.add_collection3d(faces)
    
        self.axis.scatter(vertexes[:, 0], vertexes[:, 1], vertexes[:, 2], s=0)
        if label and self.use_labels: self.axis.text(center[0], center[1], center[2], label)

        centers[oe_index, :] = center
        limits[oe_index, 0, :] = numpy.array([numpy.min(vertexes[:, 0]), numpy.max(vertexes[:, 0])])
        limits[oe_index, 1, :] = numpy.array([numpy.min(vertexes[:, 1]), numpy.max(vertexes[:, 1])])
        limits[oe_index, 2, :] = numpy.array([numpy.min(vertexes[:, 2]), numpy.max(vertexes[:, 2])])

    def add_slits_filter(self, centers, limits, oe_index, distance, height, shift, aperture=None, label="Slits", aspect_ratio_modifier=AspectRatioModifier()):
        distance *= aspect_ratio_modifier.layout_reduction_factor[0]
        shift *= aspect_ratio_modifier.layout_reduction_factor[1]
        height *= aspect_ratio_modifier.layout_reduction_factor[2]

        basic_aperture = 100 / self.units_to_mm

        vertexes = [[shift - basic_aperture / 2, distance, height - basic_aperture / 2],  # surface
                    [shift + basic_aperture / 2, distance, height - basic_aperture / 2],  # surface
                    [shift - basic_aperture / 2, distance, height + basic_aperture / 2],  # surface
                    [shift + basic_aperture / 2, distance, height + basic_aperture / 2]
                    ]
        vertexes = numpy.array(vertexes)

        edges = [[vertexes[0], vertexes[1], vertexes[3], vertexes[2]]]

        faces = Poly3DCollection(edges, linewidths=1, edgecolors='k')
        faces.set_facecolor(OpticalElementsColors.SLITS)

        self.axis.add_collection3d(faces)

        self.axis.scatter(vertexes[:, 0], vertexes[:, 1], vertexes[:, 2], s=0)
        if self.use_labels:
            label = "Slits" if label is None else label
            if not (aperture is None or len(aperture) == 0): label += ": " + str(aperture[0]) + " x " + str(aperture[1])

            self.axis.text(shift, distance, height, label)

        centers[oe_index, :] = numpy.array([shift, distance, height])
        limits[oe_index, 0, :] = numpy.array([shift - basic_aperture/2, shift + basic_aperture/2])
        limits[oe_index, 1, :] = numpy.array([distance, distance])
        limits[oe_index, 2, :] = numpy.array([height - basic_aperture/2, height + basic_aperture/2])

    def add_point(self, centers, limits, oe_index, distance, height, shift, label="Sample", aspect_ratio_modifier=AspectRatioModifier()):
        distance *= aspect_ratio_modifier.layout_reduction_factor[0]
        shift *= aspect_ratio_modifier.layout_reduction_factor[1]
        height *= aspect_ratio_modifier.layout_reduction_factor[2]

        if label is None: self.axis.scatter(shift, distance, height, s=0, c='r')
        else:
            self.axis.scatter(shift, distance, height, s=30, c='g', marker='x')
            if self.use_labels: self.axis.text(shift, distance, height, label)
    
        centers[oe_index, :]   = numpy.array([shift, distance, height])
        limits[oe_index, 0, :] = numpy.array([shift, shift])
        limits[oe_index, 1, :] = numpy.array([distance, distance])
        limits[oe_index, 2, :] = numpy.array([height, height])

    def draw_radiation_lines(self, coords, color='b', rng=None):
        if not rng is None:
            cursor = numpy.where(numpy.logical_and(coords[:, 1] >= rng[0], coords[:, 1] <= rng[1]))
    
            self.axis.scatter(coords[cursor, 0], coords[cursor, 1], coords[cursor, 2], s=3, c=color)
            self.axis.plot(coords[cursor, 0][0], coords[cursor, 1][0], coords[cursor, 2][0], color=color, linewidth=1)
        else:
            self.axis.scatter(coords[:, 0], coords[:, 1], coords[:, 2], s=3, c=color)
            self.axis.plot(coords[:, 0], coords[:, 1], coords[:, 2], color=color, linewidth=1)

    def draw_central_radiation_line(self, centers, rng=None):
        self.draw_radiation_lines(centers, color='r', rng=rng)


def create_plot(figsize=(12, 7)):
    fig = Figure(figsize=figsize)
    fig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=1, hspace=1)
    fig.patch.set_facecolor('white')

    return fig, fig.add_subplot(111, projection='3d')

def initialize_arrays(number_of_elements):
    limits  = numpy.zeros((number_of_elements, 3, 2))
    centers = numpy.zeros((number_of_elements, 3))

    return centers, limits

def get_inclinations(orientation, inclination_in, inclination_out, beam_vertical_inclination, beam_horizontal_inclination):
    if orientation == Orientations.UP:
        absolute_inclination = beam_vertical_inclination + inclination_in
        beam_vertical_inclination +=  inclination_in + inclination_out
    elif orientation == Orientations.DOWN:
        absolute_inclination = beam_vertical_inclination - inclination_in
        beam_vertical_inclination -= inclination_in + inclination_out
    elif orientation == Orientations.LEFT:
        absolute_inclination = beam_horizontal_inclination + inclination_in
        beam_horizontal_inclination -= inclination_in + inclination_out
    elif orientation == Orientations.RIGHT:
        absolute_inclination = beam_horizontal_inclination - inclination_in
        beam_horizontal_inclination += inclination_in + inclination_out

    return absolute_inclination, beam_horizontal_inclination, beam_vertical_inclination

def get_height_shift(segment_to_oe, previous_height, previous_shift, beam_vertical_inclination, beam_horizontal_inclination):
    height = previous_height + segment_to_oe * numpy.sin(beam_vertical_inclination)
    shift  = previous_shift  + segment_to_oe * numpy.sin(beam_horizontal_inclination)

    return height, shift



import numpy

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QWidget, QGridLayout, QFileDialog, QMessageBox, QLabel, QComboBox, QTextEdit

from orangewidget import gui as orange_gui


current_canvas_window = None

# ----------------------------------
# Default fonts
def widgetLabel(widget, label="", labelWidth=None, **misc):
    lbl = QLabel(label, widget)
    if labelWidth:
        lbl.setFixedSize(labelWidth, lbl.sizeHint().height())
    orange_gui.miscellanea(lbl, None, widget, **misc)

    return lbl

def lineEdit(widget, master, value, label=None, labelWidth=None,
         orientation='vertical', box=None, callback=None,
         valueType=str, validator=None, controlWidth=None,
         callbackOnType=False, focusInCallback=None,
         enterPlaceholder=False, **misc):

    ledit = orange_gui.lineEdit(widget, master, value, label, labelWidth, orientation, box, callback, valueType, validator, controlWidth, callbackOnType, focusInCallback, enterPlaceholder, **misc)

    if value:
        if (valueType != str):
            ledit.setAlignment(Qt.AlignRight)

    ledit.setStyleSheet("background-color: white;")

    return ledit

def widgetBox(widget, box=None, orientation='vertical', margin=None, spacing=4, height=None, width=None, **misc):

    box = orange_gui.widgetBox(widget, box, orientation, margin, spacing, **misc)
    box.layout().setAlignment(Qt.AlignTop)

    if not height is None:
        box.setFixedHeight(height)
    if not width is None:
        box.setFixedWidth(width)

    return box

def tabWidget(widget, height=None, width=None):
    tabWidget = orange_gui.tabWidget(widget)

    if not height is None:
        tabWidget.setFixedHeight(height)
    if not width is None:
        tabWidget.setFixedWidth(width)

    tabWidget.setStyleSheet('QTabBar::tab::selected {background-color: #a6a6a6;}')

    return tabWidget

def createTabPage(tabWidget, name, widgetToAdd=None, canScroll=False, height=None, width=None, isImage=False):
    tab = orange_gui.createTabPage(tabWidget, name, widgetToAdd, canScroll)
    tab.layout().setAlignment(Qt.AlignTop)

    if not height is None:
        tab.setFixedHeight(height)
    if not width is None:
        tab.setFixedWidth(width)

    if isImage: tab.setStyleSheet("background-color: #FFFFFF;")

    return tab

def selectSaveFileFromDialog(widget, message="Save File", default_file_name="", file_extension_filter="*.*"):
    file_path = QFileDialog.getSaveFileName(widget, message, default_file_name, file_extension_filter)[0]
    if not file_path is None and not file_path.strip() == "": return file_path
    else: return None

def selectFileFromDialog(widget, previous_file_path="", message="Select File", start_directory=".", file_extension_filter="*.*"):
    file_path = QFileDialog.getOpenFileName(widget, message, start_directory, file_extension_filter)[0]
    if not file_path is None and not file_path.strip() == "": return file_path
    else: return previous_file_path

def selectDirectoryFromDialog(widget, previous_directory_path="", message="Select Directory", start_directory="."):
    directory_path = QFileDialog.getExistingDirectory(widget, message, start_directory)
    if not directory_path is None and not directory_path.strip() == "": return directory_path
    else: return previous_directory_path

def textArea(height=None, width=None, readOnly=True, noWrap=None):
        area = QTextEdit()
        area.setReadOnly(readOnly)
        area.setStyleSheet("background-color: white;")
        if noWrap is not None:
            area.setLineWrapMode(QTextEdit.NoWrap)

        if not height is None: area.setFixedHeight(height)
        if not width is None: area.setFixedWidth(width)
    
        return area
# ------------------------------------
# UTILITY CLASS
# ------------------------------------


def _set_size(dialog, width, height):
    stylesheet_string = "QLabel{"
    if not width is None:  stylesheet_string += "min-width: " + str(width) + "px;"
    if not height is None: stylesheet_string += "min-height: " + str(height) + "px;"
    stylesheet_string += "}"
    if not (width is None and height is None): dialog.setStyleSheet(stylesheet_string)

class MessageDialog(QMessageBox):
    def __init__(self, parent, message, title=None, type="information", width=None, height=None):
        super(MessageDialog, self).__init__(parent)

        self.setStandardButtons(QMessageBox.Ok)
        if type == "information": self.setIcon(QMessageBox.Information)
        elif type == "warning":   self.setIcon(QMessageBox.Warning)
        elif type == "critical":  self.setIcon(QMessageBox.Critical)
        self.setText(message)
        if title is None: self.setWindowTitle(str(type[0]).upper() + type[1:])
        else:             self.setWindowTitle(title)
        _set_size(self, width, height)

    @classmethod
    def message(cls, parent=None, message="Message", title=None, type="information", width=None, height=None):
        MessageDialog(parent, message, title, type, width, height).exec_()

class ConfirmDialog(QMessageBox):
    def __init__(self, parent, message, title, width=None, height=None):
        super(ConfirmDialog, self).__init__(parent)

        self.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
        self.setIcon(QMessageBox.Question)
        self.setText(message)
        self.setWindowTitle(title)
        _set_size(self, width, height)

    @classmethod
    def confirmed(cls, parent=None, message="Confirm Action?", title="Confirm Action", width=None, height=None):
        return ConfirmDialog(parent, message, title, width, height).exec_() == QMessageBox.Ok

class OptionDialog(QMessageBox):
    def __init__(self, parent, message, title, options, default, width=None, height=None):
        super(OptionDialog, self).__init__(parent)

        self.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
        self.setIcon(QMessageBox.Question)
        self.setText(message)
        self.setWindowTitle(title)
        _set_size(self, width, height)

        self.selection = default

        orange_gui.comboBox(widgetBox(self, "", height=40), self, label="Select Option", items=options, callback=self.set_selection, orientation="horizontal")

    def set_selection(self, index):
        self.selection = index

    @classmethod
    def get_option(cls, parent=None, message="Select Option", title="Select Option", option=["No", "Yes"], default=0, width=None, height=None):
        dlg = OptionDialog(parent, message, title, option, default, width, height)
        if dlg.exec_() == QMessageBox.Ok: return dlg.selection
        else: return None

class ValueDialog(QMessageBox):
    def __init__(self, parent, message, title, default, width=None, height=None):
        super(ValueDialog, self).__init__(parent)

        self.setStandardButtons(QMessageBox.Ok)
        self.setIcon(QMessageBox.Question)
        self.setText(message)
        self.setWindowTitle(title)
        _set_size(self, width, height)

        self.value = default

        lineEdit(widgetBox(self, "", height=40), self, "value", "", orientation="horizontal")

    @classmethod
    def get_value(cls, parent=None, message="Input Value", title="Input Option", default=0, width=None, height=None):
        dlg = ValueDialog(parent, message, title, default, width, height)
        if dlg.exec_() == QMessageBox.Ok: return dlg.value
        else: return None


from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d import art3d

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas

class FigureCanvas3D(FigureCanvas):

    def __init__(self, fig, ax, show_legend=True, show_buttons=True):
        super().__init__(fig)

        if show_buttons:
            box = widgetBox(self, "", orientation="vertical")
            orange_gui.button(box, self, "Default View", width=100, height=35, callback=self.__default_view)
            orange_gui.button(box, self, "Top View",     width=100, height=35, callback=self.__top_view)
            orange_gui.button(box, self, "Lateral View", width=100, height=35, callback=self.__lateral_view)

        self.ax = ax
        self.size_x, self.size_y = fig.get_size_inches() * fig.dpi
        self.x_c = int(self.size_x / 2)
        self.y_c = int(self.size_y / 2)

        self.last_pos_x = self.x_c
        self.last_pos_y = self.y_c

        self.__show_legend = show_legend
        self.__add_legend()

        self.mark_default_view()

    def mark_default_view(self):
        self.__initial_azim = self.ax.azim
        self.__initial_elev = self.ax.elev

    def __default_view(self):
        self.ax.view_init(azim=self.__initial_azim, elev=self.__initial_elev)
        self.draw()

    def __top_view(self):
        self.ax.view_init(azim=0.0, elev=90.0)
        self.draw()

    def __lateral_view(self):
        self.ax.view_init(azim=0.0, elev=0.0)
        self.draw()

    def __add_legend(self):
        if self.__show_legend:
            self.ax.text2D(0.05, 0.95,
                           "Mouse Left Button -> Click and Hold: Rotate\n" +  #, Double Click: Recenter\n" + \
                           "Mouse Right Button -> Click and Hold: Zoom\n" +
                           "Mouse Left & Right Buttons or Central Button -> Click and Hold: Shift",
                           transform=self.ax.transAxes,
                           color='blue')

    def __pan(self, dx, dy):
        # convert dx dy -> dxx dyy dzz
        minx, maxx, miny, maxy, minz, maxz = self.ax.get_w_lims()
        elev, azim = numpy.deg2rad(self.ax.elev), numpy.deg2rad(self.ax.azim)
        dxe = (dy / self.size_y) * numpy.sin(elev)
        dye = - (dx / self.size_x)
        dze = - (dy / self.size_y) * numpy.cos(elev)
        dxx = (maxx - minx) * (dxe * numpy.cos(azim) - dye * numpy.sin(azim))
        dyy = (maxy - miny) * (dye * numpy.cos(azim) + dxe * numpy.sin(azim))
        dzz = (maxz - minz) * (dze)
        # pan
        self.ax.set_xlim3d(minx + dxx, maxx + dxx)
        self.ax.set_ylim3d(miny + dyy, maxy + dyy)
        self.ax.set_zlim3d(minz + dzz, maxz + dzz)
        self.ax.get_proj()

    def __zoom(self, dy):
        minx, maxx, miny, maxy, minz, maxz = self.ax.get_w_lims()
        df = 1 - ((self.size_y - dy) / self.size_y)
        dx = (maxx - minx) * df
        dy = (maxy - miny) * df
        dz = (maxz - minz) * df
        self.ax.set_xlim3d(minx - dx, maxx + dx)
        self.ax.set_ylim3d(miny - dy, maxy + dy)
        self.ax.set_zlim3d(minz - dz, maxz + dz)
        self.ax.get_proj()

    def __rotate(self, dx, dy):
        self.ax.view_init(azim=art3d._norm_angle(self.ax.azim - (dx / self.size_x) * 180),
                          elev=art3d._norm_angle(self.ax.elev - (dy / self.size_y) * 180))

    def mouseMoveEvent(self, event):
        pos_x = event.pos().x() - self.x_c
        pos_y = -(event.pos().y() - self.y_c)

        dx = pos_x - self.last_pos_x
        dy = pos_y - self.last_pos_y

        if dx == 0 and dy == 0: return

        if int(event.buttons()) == 1:        self.__rotate(dx, dy) # left button
        elif int(event.buttons())==2:        self.__zoom(dy) # right button
        elif int(event.buttons()) in [3, 4]: self.__pan(dx, dy) #central button/wheel or left and right together

        self.last_pos_x = pos_x
        self.last_pos_y = pos_y

        self.draw()

    def clear_axis(self):
        self.ax.clear()
        self.__add_legend()

#######################################################################
#######################################################################
#######################################################################
# FIXING BUG ON MATPLOTLIB 2.0.0
#######################################################################
#######################################################################
#######################################################################

from silx.gui.plot.backends.BackendMatplotlib import BackendMatplotlibQt
from silx.gui.plot.PlotWindow import PlotWindow

class OasysBackendMatplotlibQt(BackendMatplotlibQt):

    def __init__(self, plot, parent=None):
        super().__init__(plot, parent)

    def _onMouseMove(self, event):
        try:
            super(OasysBackendMatplotlibQt, self)._onMouseMove(event)
        except ValueError as exception:
            if "Data has no positive values, and therefore can not be log-scaled" in str(exception):
                pass
            else:
                raise exception


def plotWindow(parent=None, backend=None,
               resetzoom=True, autoScale=True, logScale=True, grid=True,
               curveStyle=True, colormap=True,
               aspectRatio=True, yInverted=True,
               copy=True, save=True, print_=True,
               control=False, position=False,
               roi=True, mask=True, fit=False):
    if backend is None:
        backend = OasysBackendMatplotlibQt

    plot_window = PlotWindow(parent=parent, backend=backend,
                      resetzoom=resetzoom, autoScale=autoScale, logScale=logScale, grid=grid,
                      curveStyle=curveStyle, colormap=colormap,
                      aspectRatio=aspectRatio, yInverted=yInverted,
                      copy=copy, save=save, print_=print_,
                      control=control, position=position,
                      roi=roi, mask=mask, fit=fit)

    plot_window._backend.ax.ticklabel_format(axis='y', style='sci')

    return plot_window

from silx.gui.plot import ImageView, PlotToolButtons
import silx.gui.qt as qt
from silx.gui.plot.Profile import ProfileToolBar

def imageWiew(parent=None):
    image_view = ImageView(parent=parent)
    image_view._toolbar.setVisible(False)

    image_view.removeToolBar(image_view.profile)
    image_view.profile = ProfileToolBar(plot=image_view)
    image_view.addToolBar(image_view.profile)

    def _createToolBar(image_view, title, parent):
        image_view.keepDataAspectRatioButton = PlotToolButtons.AspectToolButton(parent=image_view, plot=image_view)
        image_view.keepDataAspectRatioButton.setVisible(True)

        image_view.yAxisInvertedButton = PlotToolButtons.YAxisOriginToolButton(parent=image_view, plot=image_view)
        image_view.yAxisInvertedButton.setVisible(True)

        toolbar = qt.QToolBar(title, parent)

        objects = image_view.group.actions()
        index = objects.index(image_view.colormapAction)
        objects.insert(index + 1, image_view.keepDataAspectRatioButton)
        objects.insert(index + 2, image_view.yAxisInvertedButton)

        for obj in objects:
            if isinstance(obj, qt.QAction):
                toolbar.addAction(obj)
            else:
                if obj is image_view.keepDataAspectRatioButton:
                    image_view.keepDataAspectRatioAction = toolbar.addWidget(obj)
                elif obj is image_view.yAxisInvertedButton:
                    image_view.yAxisInvertedAction = toolbar.addWidget(obj)
                else:
                    raise RuntimeError()

        return toolbar

    image_view._toolbar = _createToolBar(image_view, title='Plot', parent=image_view)
    image_view.insertToolBar(image_view._interactiveModeToolBar, image_view._toolbar)

    return image_view



import os, sys

import numpy
from PyQt5.QtCore import QRect
from PyQt5.QtWidgets import QApplication, QMessageBox

from matplotlib import cm
from oasys.widgets.gui import FigureCanvas3D
from matplotlib.figure import Figure

from orangewidget import gui
from orangewidget.settings import Setting

from oasys.widgets.widget import OWWidget
from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence

from oasys.util.oasys_objects import OasysSurfaceData

import oasys.util.oasys_util as OU

try:
    from mpl_toolkits.mplot3d import Axes3D  # necessario per caricare i plot 3D
except:
    pass

class OWSurfaceFileReader(OWWidget):
    name = "Surface File Reader"
    id = "surface_file_reader"
    description = "Surface File Reader"
    icon = "icons/surface_reader.png"
    author = "Luca Rebuffi"
    maintainer_email = "lrebuffi@anl.gov"
    priority = 3
    category = ""
    keywords = ["surface_file_reader"]

    outputs = [{"name": "Surface Data",
                "type": OasysSurfaceData,
                "doc": "Surface Data",
                "id": "Surface Data"}]

    want_main_area = 1
    want_control_area = 1

    MAX_WIDTH = 1320
    MAX_HEIGHT = 700

    IMAGE_WIDTH = 860
    IMAGE_HEIGHT = 645

    CONTROL_AREA_WIDTH = 405
    TABS_AREA_HEIGHT = 618

    xx = None
    yy = None
    zz = None

    surface_file_name = Setting('surface.hdf5')

    negate = Setting(0)

    def __init__(self):
        super().__init__()

        geom = QApplication.desktop().availableGeometry()
        self.setGeometry(QRect(round(geom.width() * 0.05),
                               round(geom.height() * 0.05),
                               round(min(geom.width() * 0.98, self.MAX_WIDTH)),
                               round(min(geom.height() * 0.95, self.MAX_HEIGHT))))

        self.setMaximumHeight(self.geometry().height())
        self.setMaximumWidth(self.geometry().width())

        gui.separator(self.controlArea)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Read Surface", callback=self.read_surface)
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Render Surface", callback=self.render_surface)
        button.setFixedHeight(45)

        input_box_l = oasysgui.widgetBox(self.controlArea, "Input", addSpace=True, orientation="vertical", height=self.TABS_AREA_HEIGHT)

        box = oasysgui.widgetBox(input_box_l, "", addSpace=False, orientation="horizontal")

        self.le_surface_file_name = oasysgui.lineEdit(box, self, "surface_file_name", "Surface File Name",
                                                        labelWidth=120, valueType=str, orientation="horizontal")

        gui.button(box, self, "...", callback=self.selectSurfaceFile)

        gui.comboBox(input_box_l, self, "negate", label="Invert Surface", labelWidth=350,
                     items=["No", "Yes"], sendSelectedValue=False, orientation="horizontal")

        self.figure = Figure(figsize=(600, 600))
        self.figure.patch.set_facecolor('white')

        self.axis = self.figure.add_subplot(111, projection='3d')

        self.axis.set_zlabel("Z [m]")

        self.figure_canvas = FigureCanvas3D(ax=self.axis, fig=self.figure)

        self.mainArea.layout().addWidget(self.figure_canvas)

        gui.rubber(self.mainArea)

    def read_surface(self):
        try:
            self.surface_file_name = congruence.checkDir(self.surface_file_name)

            xx, yy, zz = OU.read_surface_file(self.surface_file_name)
            zz = zz if self.negate==0 else -1.0*zz

            self.xx = xx
            self.yy = yy
            self.zz = zz

            self.send("Surface Data", OasysSurfaceData(xx=self.xx, yy=self.yy, zz=self.zz, surface_data_file=self.surface_file_name))
        except Exception as exception:
            QMessageBox.critical(self, "Error",
                                 exception.args[0],
                                 QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception

    def render_surface(self):
        try:
            self.surface_file_name = congruence.checkDir(self.surface_file_name)

            xx, yy, zz = OU.read_surface_file(self.surface_file_name)
            zz = zz if self.negate==0 else -1.0*zz

            self.xx = xx
            self.yy = yy
            self.zz = zz

            self.axis.clear()

            x_to_plot, y_to_plot = numpy.meshgrid(self.xx, self.yy)

            self.axis.plot_surface(x_to_plot, y_to_plot, zz,
                                   rstride=1, cstride=1, cmap=cm.autumn, linewidth=0.5, antialiased=True)

            self.axis.set_xlabel("X [m]")
            self.axis.set_ylabel("Y [m]")
            self.axis.set_zlabel("Z [m]")
            self.axis.mouse_init()

            self.figure_canvas.draw()

            self.send("Surface Data", OasysSurfaceData(xx=self.xx, yy=self.yy, zz=self.zz, surface_data_file=self.surface_file_name))

        except Exception as exception:
            QMessageBox.critical(self, "Error",
                                 exception.args[0],
                                 QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception

    def selectSurfaceFile(self):
        self.le_surface_file_name.setText(oasysgui.selectFileFromDialog(self, self.surface_file_name, "Select Input File", file_extension_filter="HDF5 Files (*.hdf5 *.h5 *.hdf)"))





class OasysSurfaceData(object):
    def __init__(self,
                 xx=None,
                 yy=None,
                 zz=None,
                 surface_data_file=None):
        self.xx = xx
        self.yy = yy
        self.zz = zz
        self.surface_data_file=surface_data_file

class OasysErrorProfileData(object):

    def __init__(self,
                 surface_data=None,
                 error_profile_x_dim=0.0,
                 error_profile_y_dim=0.0):
        self.surface_data = surface_data
        self.error_profile_x_dim = error_profile_x_dim
        self.error_profile_y_dim = error_profile_y_dim

class OasysPreProcessorData(object):

    def __init__(self, error_profile_data=None, reflectivity_data=None):
        super().__init__()

        self.error_profile_data = error_profile_data
        self.reflectivity_data = reflectivity_data
        self.additional_data = None

    def set_additional_data(self, key, value):
        if self._additional_data is None:
            self._additional_data = {key : value}
        else:
            self._additional_data[key] = value

    def get_additional_data(self, key):
        return self._additional_data[key]

    def has_additional_data(self, key):
        return key in self._additional_data

class OasysThicknessErrorsData(object):
    def __init__(self, thickness_error_profile_data_files=[]):
        self.thickness_error_profile_data_files = thickness_error_profile_data_files



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2022, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2022. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #

class AlreadyInitializedError(ValueError):
    def __init__(self, message=None): super(AlreadyInitializedError, self).__init__(message)

class GenericRegistry(object):
    _NO_APPLICATION = "<NO APPLICATION>"

    def __init__(self, registry_name):
        self.__registry_name = registry_name
        self.__registry = {self._NO_APPLICATION: None}

    def register_instance(self, instance, application_name=None, replace=False):
        if instance is None: raise ValueError(self.__registry_name + " Instance is None")

        application_name = self.__get_application_name(application_name)

        if application_name in self.__registry.keys():
            if self.__registry[application_name] is None or replace==True: self.__registry[application_name] = instance
            else: raise AlreadyInitializedError(self.__registry_name + " Instance already initialized")
        else: self.__registry[application_name] = instance

    def reset(self, application_name=None):
        application_name = self.__get_application_name(application_name)

        if application_name in self.__registry.keys(): self.__registry[self.__get_application_name(application_name)] = None
        else: raise ValueError(self.__registry_name + " Instance not existing")

    def get_instance(self, application_name=None):
        application_name = self.__get_application_name(application_name)

        if application_name in self.__registry.keys(): return self.__registry[self.__get_application_name(application_name)]
        else: raise ValueError(self.__registry_name + " Instance not existing")

    def __get_application_name(self, application_name):
        return self._NO_APPLICATION if application_name is None else application_name



from PyQt5.QtWidgets import QApplication
from PyQt5 import QtGui, QtWidgets
from PyQt5.QtWidgets import QFileDialog, QMessageBox
from PyQt5.QtCore import QRect

from orangewidget import gui,widget
from orangewidget.settings import Setting
from oasys.widgets import gui as oasysgui, congruence
from oasys.widgets import widget as oasyswidget

import urllib.request
from urllib.request import urlretrieve

import re

def get_usual_servers():
    urls = [
        "",
        "https://github.com/oasys-kit/ShadowOui-tutorial",
        "https://github.com/oasys-kit/ShadowOui-Tutorial/tree/master/OTHER_EXAMPLES",
        "https://github.com/oasys-kit/ShadowOui-Tutorial/tree/master/SOS-WORKSHOP/EBS-WIGGLERS",
        "https://github.com/oasys-kit/oasys_school/tree/master/second/session_xoppy",
        "https://github.com/oasys-kit/oasys_school/tree/master/second/session_shadowoui",
        "https://github.com/oasys-kit/oasys_school/tree/master/second/session_srw",
        "https://github.com/oasys-esrf-kit/oasys_hercules/tree/main/2022",
        "https://github.com/oasys-esrf-kit/OasysWorkspaces",
        "https://github.com/PaNOSC-ViNYL/Oasys-PaNOSC-Workspaces",
        "https://github.com/oasys-esrf-kit/paper-multioptics-resources/tree/main/workspaces",
        "https://github.com/oasys-als-kit/Paper_JSR_yi5097/tree/master/Oasys",
        "https://github.com/oasys-als-kit/Paper_JSR_hf5419",
        "https://github.com/srio/paper-hierarchical-resources/tree/master/Oasys",
    ]
    names = [
        "",
        "ShadowOui-tutorial",
        "ShadowOui-tutorial more",
        "ShadowOui-tutorial ESRF wigglers",
        "Oasys school: xoppy",
        "Oasys school: shadowOui",
        "Oasys school: SRW",
        "Oasys HERCULES school",
        "ESRF Oasys repo",
        "PaNOSC Oasys repo",
        "Paper multioptics",
        "Paper heat load compensation",
        "Paper diaboloid",
        "Paper hierarchical",
    ]
    return names, urls

def ls_ows(target_url = "https://github.com/PaNOSC-ViNYL/Oasys-PaNOSC-Workspaces"):
    data = urllib.request.urlopen(target_url)
    list_files = []
    for line in data:
        line1 = str(line)

        if ".ows" in line1:
            try:
                found = re.search('ows">(.+?)</a>', line1).group(1)
                list_files.append(found)
            except:
                pass

    return list_files

def get_raw_dir(target_url):

    if "tree" in target_url:
        # https://github.com/oasys-kit/ShadowOui-Tutorial/tree/master/OTHER_EXAMPLES
        # https://raw.githubusercontent.com/oasys-kit/ShadowOui-Tutorial/master/OTHER_EXAMPLES/CRL_Snigirev_1996.ows
        return "https://raw.githubusercontent.com/" + target_url[19:].replace("tree/","") + "/"
    else:
        # https://raw.githubusercontent.com/PaNOSC-ViNYL/Oasys-PaNOSC-Workspaces/master/EBS_ID18.ows
        # "https://raw.githubusercontent.com/PaNOSC-ViNYL/Oasys-PaNOSC-Workspaces/master/EBS_ID32.ows"
        return "https://raw.githubusercontent.com/"+target_url[19:]+"/master/"

def get_ows_data(url, debug=0):
    url_raw = get_raw_dir(url)
    list_files = ls_ows(url)
    titles = []
    descriptions = []
    url_files1 = []
    for file in list_files:
        url_file = url_raw + file
        url_files1.append(url_file)
        if debug: print(">>", url_file)
        try:
            data = urllib.request.urlopen(url_file).read(2000).decode('utf-8')
        except:
            raise ("Error reading data from ", url_file)

        try:
            description = re.search('<scheme description="(.+?)" title', data).group(1)
            descriptions.append(description)
        except:
            descriptions.append("")

        try:
            title = re.search('title="(.+?)"', data).group(1)
            titles.append(title)
        except:
            titles. append("")


    return list_files, url_files1, titles, descriptions

class OWRemoteGithubDownloader(oasyswidget.OWWidget):
    name = "Remote Github Repository Downloader"
    description = "Utility: Remote Github Repository Downloader"
    icon = "icons/github.png"
    maintainer = "Aljosa Hafner and Manuel Sanchez del Rio"
    maintainer_email = "srio(@at@)esrf.eu"
    priority = 13
    category = "Utility"
    keywords = ["remote", "repository", "load", "read", "beamline"]

    want_main_area = 0
    beam_file_name = Setting("")
    selectedIndex = Setting([0])
    selectedFile = Setting("")
    servers_list = Setting(1)
    repository = Setting("https://github.com/oasys-kit/ShadowOui-tutorial")
    # repository = Setting("https://github.com/PaNOSC-ViNYL/Oasys-PaNOSC-Workspaces")
    urlselectedfile = Setting("")

    beamlines = []    # workspace names
    descriptions = [] # as stored in the workspace
    url_files = []    # the raw url to open remote



    def __init__(self):
        super().__init__()

        """Browse recent schemes. Return QDialog.Rejected if the user
        canceled the operation and QDialog.Accepted otherwise.

        """


        wWidth = 650
        wHeight = 400


        geom = QApplication.desktop().availableGeometry()
        self.setGeometry(QRect(round(geom.width()*0.05),
                               round(geom.height()*0.05),
                               round(min(geom.width()*0.98, wWidth)),
                               round(min(geom.height()*0.95, wHeight))))

        self.setMaximumHeight(self.geometry().height())
        self.setMaximumWidth(self.geometry().width())

        main_box = oasysgui.widgetBox(self.controlArea, "", orientation="vertical", width=wWidth-40, height=wHeight-30)

        gui.comboBox(main_box, self, "servers_list", label="Select a server",
                     items=get_usual_servers()[0], labelWidth=300,
                     sendSelectedValue=False, orientation="horizontal", callback=self.set_server)

        box1 = gui.widgetBox(main_box, orientation="horizontal")
        self.le_repoName = oasysgui.lineEdit(box1, self, "repository", "Repository URL: ", labelWidth=120,
                                             valueType=str, orientation="horizontal", callback=self.change_repo_url,
                                             )

        box2 = gui.widgetBox(main_box, orientation="horizontal")
        upper_box = oasysgui.widgetBox(box2, "", orientation="horizontal", width=wWidth-50, height=wHeight-40)

        left_box = oasysgui.widgetBox(upper_box, "List of workspaces", orientation="vertical",
                                        width=wWidth//2-13, height=wHeight-143)

        self.beamlineList = gui.listBox(left_box, self, "selectedIndex", callback=self.select_beamline)

        right_box = oasysgui.widgetBox(upper_box, "Description", addSpace=True,
                                        orientation="vertical",)
                                        # width=wWidth/2.-12., height=wHeight-143)

        self.metadataLabel = ""
        self.md_label = "%(metadataLabel)s"
        self.box_metaData = gui.label(right_box, self, self.md_label, orientation="vertical")

        box3 = gui.widgetBox(main_box, orientation="vertical")
        box33 = gui.widgetBox(box3, orientation="horizontal")
        oasysgui.lineEdit(box33, self, "urlselectedfile", "Selected file URL: ", labelWidth=120,
                                             valueType=str, orientation="horizontal",
                                             )

        box34 = gui.widgetBox(box3, orientation="horizontal")
        button = gui.button(box34, self, "Download...", callback=self.download_scheme)
        button.setFixedHeight(35)

        # button = gui.button(box34, self, "Open...", callback=self.open_scheme)
        # button.setFixedHeight(35)

        gui.rubber(self.controlArea)

        self.change_repo_url()

    def change_repo_url(self):

        self.clear_repo_url()

        try:
            beamlines, url_files, titles, descriptions = get_ows_data(self.repository)
        except:
            QMessageBox.information(self, "Error", "Error retrieving files from " + self.repository, QMessageBox.Ok)
            return

        self.descriptions = descriptions
        self.url_files = url_files
        self.beamlines = beamlines

        for i, beamline in enumerate(beamlines):
            currentName = beamline
            self.beamlineList.insertItem(i, currentName)

    def clear_repo_url(self):
        self.beamlineList.clear()
        self.selectedIndex = [0]
        self.urlselectedfile = ""


    def select_beamline(self):
        if len(self.selectedIndex) > 0:
            self.metadataLabel   = self.descriptions[self.selectedIndex[0]]
            self.selectedFile = self.beamlines[self.selectedIndex[0]]
            self.urlselectedfile = self.url_files[self.selectedIndex[0]]
        else:
            self.metadataLabel   = ""
            self.selectedFile    = ""
            self.urlselectedfile = ""



    def download_scheme(self):
        if self.selectedFile == "":
            QMessageBox.information(self, "Error", "Please select a file", QMessageBox.Ok)
            return

        filedialog = QtWidgets.QFileDialog(self)
        filename, _ = filedialog.getSaveFileName(self, "Save workspace", self.selectedFile, "Oasys files (*.ows)")

        if filename == "":
            return

        try:
            filepath, http_msg = urlretrieve(self.urlselectedfile,
                                             filename=filename,
                                             reporthook=None,
                                             data=None)
            print("Workspace file %s downloaded to %s" % (self.urlselectedfile, filepath))
            QMessageBox.information(self, "Success", "File saved: " + filepath, QMessageBox.Ok)

        except IOError:
            QMessageBox.information(self, "Error", "Error writing file: " + filename, QMessageBox.Ok)

    def set_server(self):
        names, urls = get_usual_servers()

        self.repository = urls[self.servers_list]
        self.change_repo_url()



if __name__ == "__main__":

    if False:
        names, servers = get_usual_servers()
        print("names:", names)
        print("servers:", servers)
        for server in servers:
            print("")
            print(">>>>>>>>>>>>>>>>>server: ", server)
            if server != "":
                beamlines, url_files, titles, descriptions = get_ows_data(server)
                print("          beamlines: ", beamlines)
                print("          url_files: ", url_files)
                print("          titles: ", titles)
                print("          descriptions: ", descriptions)

    import sys
    a = QApplication(sys.argv)
    ow = OWRemoteGithubDownloader()
    ow.show()
    a.exec_()
    ow.saveSettings()





from unittest.mock import Mock
import pytest
from oasys.application.addons import is_updatable, Available, Installed, Installable


@pytest.mark.parametrize('items,expected', [
    (Available(installable=True), False),
    (Installed(installable=None, local=Mock()), False),
    (Installed(installable=Installable(name='mock_name',
                                       version='2.0.0',
                                       summary='mock_summary',
                                       description='mock_description',
                                       package_url='mock_package_url',
                                       release_urls='mock_release_urls'),
               local=Mock(version='1.0.0')), True),
])
def test_is_updatable(items, expected):
    assert is_updatable(items) == expected



__author__ = 'labx'

"""
=========
Loop Management
=========

Widgets for loop management

"""

# Category description for the widget registry

NAME = "Basic Loops"

DESCRIPTION = "Widgets for Loop Management."

BACKGROUND = "#FAAC58"

ICON = "icons/loop_management.png"

PRIORITY = 2
